package ru.dbotthepony.kstarbound

import org.apache.logging.log4j.LogManager
import org.lwjgl.Version
import org.lwjgl.glfw.GLFW.glfwSetWindowShouldClose
import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.shapes.ChainShape
import ru.dbotthepony.kbox2d.collision.shapes.CircleShape
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kbox2d.dynamics.B2World
import ru.dbotthepony.kbox2d.dynamics.Body
import ru.dbotthepony.kbox2d.dynamics.joint.MouseJoint
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.entities.Move
import ru.dbotthepony.kstarbound.world.entities.PlayerEntity
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import java.io.File
import java.util.*
import kotlin.collections.ArrayList

private val LOGGER = LogManager.getLogger()

fun main() {
	LOGGER.info("Running LWJGL ${Version.getVersion()}")

	if (true) {
		//val pak = StarboundPak(File("J:\\SteamLibrary\\steamapps\\common\\Starbound\\assets\\packed.pak"))
		//val json = JsonParser.parseReader(pak.getReader("/projectiles/traps/lowgravboostergas/lowgravboostergas.projectile"))
		//val obj = Gson().fromJson(json, ProjectileDefinitionBuilder::class.java)
		//println(obj.build())
		//return
	}

	val world = B2World(Vector2d(y = -10.0))

	val groundDef = BodyDef(
		position = Vector2d(y = 0.0)
	)

	val ground = world.createBody(groundDef)

	//val groundPoly = PolygonShape()
	//groundPoly.setAsBox(50.0, 10.0)

	val groundPoly = ChainShape()
	groundPoly.createLoop(listOf(
		Vector2d(-30.0, 10.0),
		Vector2d(-25.0, 0.0),
		Vector2d(25.0, 0.0),
		Vector2d(30.0, 10.0),
		Vector2d(30.0, -2.0),
		Vector2d(-30.0, -2.0)).asReversed())

	ground.createFixture(groundPoly, 0.0)

	val boxes = ArrayList<Body>()

	/*run {
		val movingDef = BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(y = 4.0),
			angle = PI / 4.0
		)
		val movingBody = world.createBody(movingDef)

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		movingBody.createFixture(fixtureDef)
		boxes.add(movingBody)
	}*/

	run {
		if (true)
			return@run
		val movingDef = BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -1.0, y = 6.0),
		)

		val movingBody = world.createBody(movingDef)

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		movingBody.createFixture(fixtureDef)
		boxes.add(movingBody)
	}

	val rand = Random()

	val mouseJoints = ArrayList<MouseJoint>()

	run {
		val stripes = 4

		for (stripe in 0 until stripes) {
			for (x in 0 .. (stripes - stripe)) {
				val movingBody = world.createBody(BodyDef(
					type = BodyType.DYNAMIC,
					position = Vector2d(x = (-stripes + stripe) * 1.0 + x * 2.1, y = 8.0 + stripe * 2.1),
					gravityScale = 1.1
				))

				val dynamicBox: IShape<*>

				if (false) {
					dynamicBox = PolygonShape()
					dynamicBox.setAsBox(1.0, 1.0)
				} else {
					dynamicBox = CircleShape(1.0)
				}

				movingBody.createFixture(FixtureDef(
					shape = dynamicBox,
					density = 1.0,
					friction = 0.3
				))

				/*for (otherBody in boxes) {
					val def = DistanceJointDef(movingBody, otherBody, movingBody.position, otherBody.position)
					world.createJoint(def)
				}*/

				/*val mouse = world.createJoint(MouseJointDef(
					Vector2d(y = 10.0, x = 10.0),
					bodyB = movingBody,
					maxForce = 1000.0
				).linearStiffness(50.0, 0.7, bodyB = movingBody, bodyA = null))

				mouseJoints.add(mouse as MouseJoint)*/

				boxes.add(movingBody)
			}
		}
	}

	run {
		if (true)
			return@run
		val movingDef1 = BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -4.0, y = 4.0),
			gravityScale = 1.1
		)

		val movingBody1 = world.createBody(movingDef1)

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef1 = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		movingBody1.createFixture(fixtureDef1)

		boxes.add(movingBody1)

		val movingDef2 = BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 4.0, y = 6.0),
			gravityScale = 1.1
		)

		val movingBody2 = world.createBody(movingDef2)

		val fixtureDef2 = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		movingBody2.createFixture(fixtureDef2)
		boxes.add(movingBody2)

		val groundAnchor1 = movingBody1.position + Vector2d(y = 10.0)
		val groundAnchor2 = movingBody2.position + Vector2d(y = 12.0)
		val def = PulleyJointDef(movingBody1, movingBody2, groundAnchor1, groundAnchor2, movingBody1.position, movingBody2.position, 1.0)

		world.createJoint(def)
	}

	run {
		val base = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 0.0, y = 16.0),
		))

		val wheel1 = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -2.0, y = 15.0),
		))

		val wheel2 = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 2.0, y = 15.0),
		))

		base.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(2.0, 0.5) },
			density = 1.0
		))

		wheel1.createFixture(FixtureDef(
			shape = CircleShape(0.5),
			density = 1.0,
			friction = 1.5
		))

		wheel2.createFixture(FixtureDef(
			shape = CircleShape(0.5),
			density = 1.0,
			friction = 1.5
		))

		world.createJoint(WheelJointDef(
			bodyA = base,
			bodyB = wheel1,
			anchor = Vector2d(x = -2.0, y = 15.0),
			axis = Vector2d.POSITIVE_Y,
			enableLimit = true,
			upperTranslation = 0.25,
			lowerTranslation = -0.25,
		))

		world.createJoint(WheelJointDef(
			bodyA = base,
			bodyB = wheel2,
			anchor = Vector2d(x = 2.0, y = 15.0),
			axis = Vector2d.POSITIVE_Y,
			enableLimit = true,
			upperTranslation = 0.25,
			lowerTranslation = -0.25,
		))

		base.setTransform(base.position, 0.56)
	}

	run {
		val circleNail = world.createBody(BodyDef(
			type = BodyType.STATIC,
			position = Vector2d(x = -12.0, y = 4.0),
		))

		val circleBody = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -12.0, y = 4.0),
		))

		circleBody.createFixture(FixtureDef(
			shape = CircleShape(2.0),
			density = 1.0,
			friction = 0.3
		))

		val circleNailJoint = world.createJoint(RevoluteJointDef(
			circleNail,
			circleBody,
			Vector2d(x = -12.0, y = 4.0)
		))

		val boxNail = world.createBody(BodyDef(
			type = BodyType.STATIC,
			position = Vector2d(x = 2.0, y = 4.0),
		))

		val boxBody = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 2.0, y = 4.0),
		))

		boxBody.createFixture(FixtureDef(
			shape = PolygonShape().also { it.setAsBox(3.5, 0.5) },
			density = 1.0,
			friction = 0.3
		))

		val boxNailJoint = world.createJoint(RevoluteJointDef(
			boxNail,
			boxBody,
			Vector2d(x = 2.0, y = 4.0)
		))

		val gearJoint = world.createJoint(GearJointDef(
			circleBody,
			boxBody,
			circleNailJoint,
			boxNailJoint,
			1.0
		))
	}

	run {
		val boxA = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -2.0, y = 6.0),
		))

		val boxB = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 0.0, y = 6.0),
		))

		val boxC = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = 2.0, y = 6.0),
		))

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		boxA.createFixture(fixtureDef)
		boxB.createFixture(fixtureDef)
		boxC.createFixture(fixtureDef)

		world.createJoint(WeldJointDef(
			bodyA = boxA, bodyB = boxB, (boxA.position + boxB.position) * 0.5
		).linearStiffness(5.0, 0.5, boxA, boxB))

		world.createJoint(WeldJointDef(
			bodyA = boxB, bodyB = boxC, (boxB.position + boxC.position) * 0.5
		).linearStiffness(5.0, 0.5, boxB, boxC))
	}

	run {
		val boxA = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -2.0, y = 8.0),
		))

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		boxA.createFixture(fixtureDef)

		world.createJoint(FrictionJointDef(
			bodyA = boxA,
			bodyB = ground,
			Vector2d(x = -2.0, y = 8.0),
			maxForce = 1.0,
			maxTorque = 10.0,
			collideConnected = true
		))
	}

	run {
		val boxA = world.createBody(BodyDef(
			type = BodyType.DYNAMIC,
			position = Vector2d(x = -2.0, y = 8.0),
		))

		val dynamicBox = PolygonShape()
		dynamicBox.setAsBox(1.0, 1.0)

		val fixtureDef = FixtureDef(
			shape = dynamicBox,
			density = 1.0,
			friction = 0.3
		)

		boxA.createFixture(fixtureDef)

		world.createJoint(MotorJointDef(
			bodyA = boxA,
			bodyB = ground,
			maxForce = 1000.0,
			collideConnected = true,
		))
	}

	val timeStep = 1.0 / 144.0

	val client = StarboundClient()

	//Starbound.addFilePath(File("./unpacked_assets/"))
	Starbound.addPakPath(File("J:\\SteamLibrary\\steamapps\\common\\Starbound\\assets\\packed.pak"))

	/*Starbound.initializeGame { finished, replaceStatus, status ->
		client.putDebugLog(status, replaceStatus)
	}*/

	client.onTermination {
		Starbound.terminateLoading = true
	}

	var chunkA: Chunk<*, *>? = null

	val ent = PlayerEntity(client.world!!)

	Starbound.onInitialize {
		/*chunkA = client.world!!.computeIfAbsent(ChunkPos(0, 0)).chunk
		val chunkB = client.world!!.computeIfAbsent(ChunkPos(-1, 0)).chunk
		val chunkC = client.world!!.computeIfAbsent(ChunkPos(-2, 0)).chunk

		val tile = Starbound.getTileDefinition("alienrock")

		for (x in -48 .. 48) {
			for (y in 0 .. 20) {
				val chnk = client.world!!.computeIfAbsent(ChunkPos(x, y))

				if (y == 0) {
					for (bx in 0 .. 31) {
						for (by in 0 .. 3) {
							chnk.chunk.foreground[bx, by] = tile
						}
					}
				}
			}
		}

		run {
			val chunk = client.world!!.computeIfAbsent(ChunkPos(-2, 0)).chunk

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0 .. (CHUNK_SIZE_FF - y)) {
					chunk.foreground[x, y] = tile
				}
			}
		}

		run {
			val chunk = client.world!!.computeIfAbsent(ChunkPos(-3, 0)).chunk

			for (y in 0 .. CHUNK_SIZE_FF) {
				for (x in 0 .. (CHUNK_SIZE_FF - y * 2)) {
					chunk.foreground[x, y] = tile
				}
			}
		}

		for (x in 0 .. 31) {
			for (y in 0 .. 3) {
				chunkA!!.foreground[x, y] = tile
			}
		}

		for (x in 0 .. 31) {
			for (y in 8 .. 9) {
				chunkA!!.foreground[x, y] = tile
			}
		}

		for (x in 0 .. 31) {
			for (y in 0 .. 0) {
				chunkB.foreground[x, y] = tile
			}
		}

		for (x in 4 .. 8) {
			for (y in 4 .. 8) {
				chunkA!!.foreground[x, y] = null as TileDefinition?
			}
		}

		/*val rand = Random()

		for (i in 0 .. 400) {
			chunkA!!.foreground[rand.nextInt(0, CHUNK_SIZE_FF), rand.nextInt(0, CHUNK_SIZE_FF)] = tile
		}*/

		ent.movement.dropToFloor()

		for ((i, proj) in Starbound.projectilesAccess.values.withIndex()) {
			val projEnt = Projectile(client.world!!, proj)
			projEnt.pos = Vector2d(i * 2.0, 10.0)
			projEnt.spawn()
		}*/
	}

	ent.pos += Vector2d(y = 36.0, x = -10.0)

	client.onDrawGUI {
		client.gl.font.render("${ent.pos}", y = 100f, scale = 0.25f)
		client.gl.font.render("${ent.movement.velocity}", y = 120f, scale = 0.25f)
	}

	client.onPreDrawWorld {
		//client.camera.pos.x = ent.pos.x.toFloat()
		//client.camera.pos.y = ent.pos.y.toFloat()
	}

	client.camera.pos.y = 10f

	world.debugDraw = client.gl.box2dRenderer

	client.gl.box2dRenderer.drawShapes = true
	client.gl.box2dRenderer.drawPairs = false
	client.gl.box2dRenderer.drawAABB = false
	client.gl.box2dRenderer.drawJoints = false

	client.onPostDrawWorld {
		world.debugDraw()

		client.gl.quadWireframe {
			var pos: Vector2d

			/*for (box in boxes) {
				pos = box.position

				it.quadRotated(
					-1f,
					-1f,
					1f,
					1f,
					pos.x.toFloat(),
					pos.y.toFloat(),
					box.angle,
				)
			}*/

			/*for (box in boxes) {
				val broad = world.contactManager.broadPhase
				val f = box.fixtureList!!
				val proxy = f.proxies[0].proxyId
				val aabb = broad.getFatAABB(proxy)

				it.quad(
					aabb,
				)
			}*/

			/*pos = ground.position

			it.quad(
				(pos.x - 50f).toFloat(),
				(pos.y - 10f).toFloat(),
				(pos.x + 50f).toFloat(),
				(pos.y + 10f).toFloat(),
			)*/
		}

		for (joint in mouseJoints) {
			joint.targetA = Vector2d(rand.nextDouble() * 20.0, rand.nextDouble() * 20.0)

			if (rand.nextDouble() < 0.01) {
				world.destroyJoint(joint)
				mouseJoints.remove(joint)
				break
			}
		}

		world.step(timeStep, 6, 4)
	}

	ent.spawn()

	while (client.renderFrame()) {
		Starbound.pollCallbacks()

		//ent.think(client.frameRenderTime)
		//client.camera.pos.x = ent.pos.x.toFloat()
		//client.camera.pos.y = ent.pos.y.toFloat()

		//println(client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1)

		//if (ent.onGround)
			//ent.velocity += client.camera.velocity.toDoubleVector() * client.frameRenderTime * 0.1

		if (client.input.KEY_LEFT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_LEFT
		} else if (client.input.KEY_RIGHT_DOWN) {
			ent.movement.moveDirection = Move.MOVE_RIGHT
		} else {
			ent.movement.moveDirection = Move.STAND_STILL
		}

		if (client.input.KEY_SPACE_PRESSED && ent.movement.onGround) {
			ent.movement.requestJump()
		} else if (client.input.KEY_SPACE_RELEASED) {
			ent.movement.recallJump()
		}

		if (client.input.KEY_ESCAPE_PRESSED) {
			glfwSetWindowShouldClose(client.window, true)
		}

		ent.wantsToDuck = client.input.KEY_DOWN_DOWN

		//if (chunkA != null && glfwGetTime() < 10.0) {
		//	val tile = Starbound.getTileDefinition("alienrock")
			//chunkA!!.foreground[rand.nextInt(0, CHUNK_SIZE_FF), rand.nextInt(0, CHUNK_SIZE_FF)] = tile
		//}
	}
}
